import abc
from abc import ABC

from NEMO.models import Tool
from django.utils.timezone import now

from NEMO_fabublox_integration.models import FabubloxIntegrationConfiguration, ToolField


class Serializer(ABC):

    def __init__(self, instance_id: str, type_name: str):
        self.entity_type = type_name
        self.instance_id = instance_id
        self.instance_type = "NEMO"

    def get_obj_id(self, obj):
        return str(obj.id)

    def serialize(self, obj, fields=None):
        return {
            "source_type": self.instance_type,
            "source_id": self.instance_id,
            "entity_type": self.entity_type,
            "source_entity_id": self.get_obj_id(obj),
            "file_date_created": now().isoformat(),
            "data": self.serialize_entity(obj, fields),
        }

    def get_object_path(self, obj):
        timestamp = str(int(now().timestamp() * 1000))
        return f"{self.instance_type}/{self.instance_id}/{self.entity_type}/{self.get_obj_id(obj)}/{timestamp}.json"

    @abc.abstractmethod
    def serialize_entity(self, obj, fields: list[str] = None):
        return {}


class ToolSerializer(Serializer):
    def __init__(self, instance_id: str):
        super().__init__(instance_id, "tool")

    def serialize_entity(self, obj: Tool, fields: list[str] = None):
        # Name is always required
        data: dict = {"name": obj.name}
        if fields is None:
            fields = []

        if ToolField.DESCRIPTION.value in fields and obj.description:
            data["description"] = obj.description
        if ToolField.CATEGORY.value in fields and obj.category:
            data["category"] = obj.category
        if ToolField.LOCATION.value in fields and obj.location:
            data["location"] = obj.location

        if ToolField.PRIMARY_OWNER.value in fields and obj.primary_owner:
            primary_owner = {}
            if ToolField.PRIMARY_OWNER_FIRST_NAME.value in fields and obj.primary_owner.first_name:
                primary_owner["first_name"] = obj.primary_owner.first_name
            if ToolField.PRIMARY_OWNER_LAST_NAME.value in fields and obj.primary_owner.last_name:
                primary_owner["last_name"] = obj.primary_owner.last_name
            if ToolField.PRIMARY_OWNER_EMAIL.value in fields and obj.primary_owner.email:
                primary_owner["email"] = obj.primary_owner.email
            if len(primary_owner) > 0:
                data["primary_owner"] = primary_owner

        if ToolField.BACKUP_OWNERS.value in fields and obj.backup_owners.exists():
            backup_owners = []
            for backup_owner in obj.backup_owners.all():
                backup_owner_data = {}
                if ToolField.BACKUP_OWNERS_FIRST_NAME.value in fields and backup_owner.first_name:
                    backup_owner_data["first_name"] = backup_owner.first_name
                if ToolField.BACKUP_OWNERS_LAST_NAME.value in fields and backup_owner.last_name:
                    backup_owner_data["last_name"] = backup_owner.last_name
                if ToolField.BACKUP_OWNERS_EMAIL.value in fields and backup_owner.email:
                    backup_owner_data["email"] = backup_owner.email
                if len(backup_owner_data) > 0:
                    backup_owners.append(backup_owner_data)
            if len(backup_owners) > 0:
                data["backup_owners"] = backup_owners

        return data


def get_tool_serializer():
    lms_fabublox_instance_id = FabubloxIntegrationConfiguration.get_instance().instance_id
    if not lms_fabublox_instance_id:
        raise ValueError("Instance id for this LMS in FabuBlox is not set.")

    return ToolSerializer(lms_fabublox_instance_id)



import logging

from django.apps import AppConfig
from django.conf import settings
from . import app_settings as defaults

logger = logging.getLogger(__name__)

for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class FabubloxIntegrationConfig(AppConfig):
    name = "NEMO_fabublox_integration"
    verbose_name = "FabuBlox Integration"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        """
        This code will be run when Django starts.
        """
        from NEMO_fabublox_integration.customization import FabubloxCustomization
        from NEMO.plugins.utils import check_extra_dependencies

        check_extra_dependencies(self.name, ["NEMO", "NEMO-CE"])

        # Load the configuration model to ensure it exists
        try:
            from .models import FabubloxIntegrationConfiguration

            configuration = FabubloxIntegrationConfiguration.get_instance()
        except Exception:
            # The models may not be loaded yet (e.g., during migrations)
            return

        # check authentication key is set and is valid
        key_location = getattr(settings, "FABUBLOX_AUTHENTICATION_KEY_LOCATION", None)
        if not key_location:
            logger.warning("FABUBLOX_AUTHENTICATION_KEY_LOCATION is not set. Please set it in your settings.py file.")

        try:
            from .models import FabubloxAuthenticationKey

            if key_location:
                authentication_key = FabubloxAuthenticationKey(key_location)
                authentication_key.set_configuration_instance_id()
        except Exception as e:
            logger.error(e)

        # if the bucket name is not set, use the one from settings
        if configuration and hasattr(settings, "FABUBLOX_DATA_SYNC_BUCKET_NAME"):
            configuration.data_sync_bucket_name = settings.FABUBLOX_DATA_SYNC_BUCKET_NAME
            configuration.save()



from NEMO_fabublox_integration.views import fabublox_integration
from django.urls import path, include
from django.views.generic import RedirectView

urlpatterns = [
    path(
        "fabublox/",
        include(
            [
                path("", RedirectView.as_view(url="/fabublox/tools", permanent=False), name="fabublox_index"),
                path("configuration/", fabublox_integration.configuration, name="fabublox_configuration"),
                path(
                    "shared_tool_fields/",
                    fabublox_integration.update_shared_tool_fields,
                    name="fabublox_shared_tool_fields",
                ),
                path("tools/", fabublox_integration.tools, name="fabublox_tools"),
                path("synchronization/", fabublox_integration.synchronizations, name="fabublox_synchronization"),
                path(
                    "synchronize/tools/",
                    fabublox_integration.synchronize_tools,
                    name="fabublox_synchronize_tools",
                ),
                path(
                    "synchronize/jobs-in-progress/",
                    fabublox_integration.in_progress_synchronizations,
                    name="fabublox_in_progress_synchronization",
                ),
            ]
        ),
    )
]



from NEMO.decorators import customization
from NEMO.views.customization import CustomizationBase


@customization(key="fabublox_integration", title="FabuBlox Integration")
class FabubloxCustomization(CustomizationBase):
    variables = {}
    files = []



import json
from logging import getLogger

from google.cloud import storage
from google.oauth2 import service_account

gcs_logger = getLogger(__name__)


def get_gcs_bucket(sa_key, bucket_name):
    try:
        credentials = service_account.Credentials.from_service_account_info(sa_key)
        client = storage.Client(credentials=credentials, project=credentials.project_id)
        return client.bucket(bucket_name)
    except Exception as e:
        gcs_logger.error(f"Failed to initialize GCS bucket '{bucket_name}': {e}")
        raise ValueError(f"Failed to initialize GCS bucket")


def upload_json_to_gcs(data: dict, destination_blob_name: str, bucket: storage.Bucket):
    blob = bucket.blob(destination_blob_name)
    json_data = json.dumps(data)
    blob.upload_from_string(json_data, content_type="application/json")



from django.contrib import admin

from NEMO_fabublox_integration.models import FabubloxIntegrationConfiguration


@admin.register(FabubloxIntegrationConfiguration)
class FabubloxIntegrationConfigurationAdmin(admin.ModelAdmin):
    list_display = ("instance_id", "data_sync_bucket_name")
    search_fields = ("instance_id", "data_sync_bucket_name")
    exclude = ("shared_tool_fields",)
    readonly_fields = ("data_sync_bucket_name",)

    def has_add_permission(self, request):
        return FabubloxIntegrationConfiguration.objects.count() == 0

    def has_delete_permission(self, request, obj=None):
        return False



from django.apps import apps
from django.test import TestCase


class FabubloxIntegrationTest(TestCase):

    def test_plugin_is_installed(self):
        assert apps.is_installed("NEMO_fabublox_integration")



from logging import getLogger

from NEMO.decorators import postpone
from django.db import transaction
from django.db.models import F
from django.utils.timezone import now
from google.auth.exceptions import RefreshError
from google.cloud.exceptions import NotFound
from NEMO.utilities import get_model_instance

from NEMO_fabublox_integration.gcs import get_gcs_bucket, upload_json_to_gcs
from NEMO_fabublox_integration.models import (
    FabubloxSynchronizationJob,
    FabubloxSyncFailure,
    FabubloxDataSyncStatus,
    JobStatus,
    FabubloxIntegrationConfiguration,
)

synchronize_logger = getLogger(__name__)


def check_permissions(bucket):
    if not bucket:
        raise ValueError("FabuBlox access point is misconfigured.")
    try:
        # This will not check if we have access to the bucket, but it will check if the bucket exists and is accessible.
        bucket.test_iam_permissions(["storage.objects.create"])
    except NotFound:
        raise ValueError("FabuBlox access point not found or misconfigured")
    except RefreshError:
        raise ValueError("Your authentication key does not grant you permission to send data to FabuBlox.")


def synchronize_job(serializer, content_type, ids, user):
    configuration = FabubloxIntegrationConfiguration.get_instance()
    with transaction.atomic():
        if FabubloxSynchronizationJob.get_in_progress_jobs().filter(content_type=content_type).exists():
            raise ValueError("A synchronization job is already in progress. Please wait for it to complete.")
        bucket = get_gcs_bucket(
            configuration.get_fabublox_authentication_key().key, configuration.data_sync_bucket_name
        )
        check_permissions(bucket)
        job = FabubloxSynchronizationJob.create_job(content_type, ids, user)
        job.status = JobStatus.IN_PROGRESS
        job.save()
        start(bucket, serializer, content_type, ids, job.id, job.started_by, configuration.shared_tool_fields)


@postpone
def start(bucket, serializer, content_type, ids, job_id, started_by, fields=None):
    try:
        for item_id in ids:
            # Check if job is still IN_PROGRESS BEFORE doing any work
            is_in_progress = FabubloxSynchronizationJob.objects.filter(id=job_id, status=JobStatus.IN_PROGRESS).exists()
            if not is_in_progress:
                synchronize_logger.info(
                    f"Job {job_id} is no longer in progress (likely marked as interrupted), stopping synchronization thread"
                )
                return

            item_name = "Unknown"
            try:
                item = get_model_instance(content_type, item_id)
                item_name = str(item)
                data = serializer.serialize(item, fields)
                destination_blob_name = serializer.get_object_path(item)

                upload_json_to_gcs(data, destination_blob_name, bucket)

                try:
                    # Update counter
                    FabubloxSynchronizationJob.objects.filter(id=job_id).update(
                        succeeded_items=F("succeeded_items") + 1, updated_at=now()
                    )

                    # Update data sync status
                    FabubloxDataSyncStatus.objects.update_or_create(
                        content_type_id=content_type.id,
                        object_id=item_id,
                        defaults={"last_synced": now(), "last_synced_by": started_by},
                    )
                except Exception as e:
                    synchronize_logger.error(
                        f"Error updating synchronization status for item {item_id} ({item_name}) for job {job_id}: {e}"
                    )

            except Exception as e:
                # Record failure
                FabubloxSyncFailure.objects.create(
                    job_id=job_id, item_id=item_id, item_name=item_name, error_message=str(e)[:1000]
                )

                # Update counter
                FabubloxSynchronizationJob.objects.filter(id=job_id).update(
                    failed_items=F("failed_items") + 1, updated_at=now()
                )

        job = FabubloxSynchronizationJob.objects.get(id=job_id)

        # Check if job was already marked as completed/failed by cleanup
        if job.status != JobStatus.IN_PROGRESS:
            return

        if job.succeeded_items == job.total_items:
            status = JobStatus.COMPLETED
        elif job.failed_items == job.total_items:
            status = JobStatus.FAILED
        else:
            status = JobStatus.PARTIALLY_COMPLETED

        job.status = status
        job.save()

    except Exception as e:
        synchronize_logger.error(f"Unexpected error in synchronization job {job_id}: {e}", exc_info=True)
        FabubloxSynchronizationJob.objects.filter(id=job_id).update(status=JobStatus.FAILED, updated_at=now())



# Generated by Django 4.2.23 on 2025-10-06 21:01

import NEMO.fields
import NEMO_fabublox_integration.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="FabubloxIntegrationConfiguration",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "instance_id",
                    models.CharField(
                        blank=True,
                        help_text="Unique identifier for this NEMO instance in Fabublox (Do not change this unless instructed by Fabublox support).",
                        max_length=255,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "data_sync_bucket_name",
                    models.CharField(
                        blank=True,
                        help_text="Name of the bucket used for data synchronization with Fabublox (Do not change this unless instructed by Fabublox support).",
                        max_length=255,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "admin_permission",
                    NEMO.fields.MultiRoleGroupPermissionChoiceField(
                        default="is_superuser",
                        help_text="The roles/groups required for users to access the Fabublox integration features.",
                    ),
                ),
                (
                    "shared_tool_fields",
                    models.JSONField(
                        blank=True,
                        default=NEMO_fabublox_integration.models.ToolField.get_all_fields,
                        help_text="List of fields to be shared with Fabublox for tool synchronization.",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="FabubloxSynchronizationJob",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "status",
                    models.IntegerField(
                        choices=[
                            (0, "Pending"),
                            (1, "In Progress"),
                            (2, "Completed"),
                            (3, "Partially Completed"),
                            (4, "Failed"),
                        ],
                        default=0,
                        help_text="Current status of the synchronization job",
                    ),
                ),
                (
                    "total_items",
                    models.PositiveIntegerField(default=0, help_text="Total number of items to be synchronized"),
                ),
                (
                    "succeeded_items",
                    models.PositiveIntegerField(default=0, help_text="Number of items successfully synchronized"),
                ),
                (
                    "failed_items",
                    models.PositiveIntegerField(default=0, help_text="Number of items that failed to synchronize"),
                ),
                ("item_ids", models.JSONField(help_text="List of item IDs being synchronized")),
                ("created_at", models.DateTimeField(auto_now_add=True, help_text="Timestamp when the job was created")),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, help_text="Timestamp when the job was last updated"),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        help_text="Content type of the model being synced",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "started_by",
                    models.ForeignKey(
                        blank=True,
                        help_text="User who started this synchronization job",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="fabublox_sync_jobs",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FabubloxSyncFailure",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("item_id", models.PositiveIntegerField(help_text="ID of the item that failed to sync")),
                ("item_name", models.CharField(help_text="Name of the item that failed", max_length=255)),
                ("error_message", models.TextField(help_text="Error message from the failed sync attempt")),
                ("failed_at", models.DateTimeField(auto_now_add=True, help_text="Timestamp when the sync failed")),
                (
                    "job",
                    models.ForeignKey(
                        help_text="Synchronization job",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="failures",
                        to="NEMO_fabublox_integration.fabubloxsynchronizationjob",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FabubloxDataSyncStatus",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("object_id", models.PositiveIntegerField()),
                (
                    "last_synced",
                    models.DateTimeField(auto_now=True, help_text="Timestamp of the last sync attempt with FabuBlox"),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        help_text="Content type of the model being synced",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "last_synced_by",
                    models.ForeignKey(
                        blank=True,
                        help_text="User who last synced this entity",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="fabublox_sync_status",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="fabubloxsynchronizationjob",
            index=models.Index(fields=["status", "-created_at"], name="NEMO_fabubl_status_9a5a01_idx"),
        ),
        migrations.AddIndex(
            model_name="fabubloxsynchronizationjob",
            index=models.Index(fields=["-created_at"], name="NEMO_fabubl_created_a9ef2a_idx"),
        ),
        migrations.AddIndex(
            model_name="fabubloxsyncfailure",
            index=models.Index(fields=["job", "-failed_at"], name="NEMO_fabubl_job_id_d062e6_idx"),
        ),
        migrations.AddIndex(
            model_name="fabubloxdatasyncstatus",
            index=models.Index(
                fields=["content_type", "object_id", "-last_synced"], name="NEMO_fabubl_content_c0441c_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="fabubloxdatasyncstatus",
            index=models.Index(fields=["-last_synced"], name="NEMO_fabubl_last_sy_323f0c_idx"),
        ),
        migrations.AlterUniqueTogether(
            name="fabubloxdatasyncstatus",
            unique_together={("content_type", "object_id")},
        ),
    ]



from django import forms

from NEMO_fabublox_integration.models import FabubloxIntegrationConfiguration


class FabubloxIntegrationSharedToolFieldsForm(forms.ModelForm):
    class Meta:
        model = FabubloxIntegrationConfiguration
        fields = ["shared_tool_fields"]



from datetime import timedelta

from NEMO.fields import MultiRoleGroupPermissionChoiceField
from NEMO.models import User
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import ForeignKey
from django.utils.timezone import now
import json
from pathlib import Path


# Global variable to track last cleanup time
# Note: In multi-process environments (e.g., gunicorn with multiple workers),
# each worker will maintain its own timestamp. This is acceptable for this use case.
_last_cleanup_time = None
INVALID_AUTHENTICATION_KEY = "FabuBlox Authentication key is not valid."
NOTFOUND_AUTHENTICATION_KEY = "FabuBlox Authentication key is not configured."


class FabubloxAuthenticationKey:
    def __init__(self, path):
        self.file_path = path
        path = Path(path)
        if not path.exists() or not path.is_file():
            raise ValueError(NOTFOUND_AUTHENTICATION_KEY)

        with path.open("r", encoding="utf-8") as file:
            file_content = file.read()
            if file_content:
                try:
                    data = json.loads(file_content)
                    content = self.read_file_metadata(data)
                    self.key = data
                    self.client_email = content["client_email"]
                    self.key_instance_id = content["key_instance_id"]

                    configuration = FabubloxIntegrationConfiguration.get_instance()
                    if configuration.instance_id and self.key_instance_id != configuration.instance_id:
                        raise ValueError(INVALID_AUTHENTICATION_KEY)

                except json.JSONDecodeError:
                    raise ValueError(INVALID_AUTHENTICATION_KEY)
            else:
                raise ValueError(NOTFOUND_AUTHENTICATION_KEY)

    @classmethod
    def read_file_metadata(cls, data):
        if (
            "type" not in data
            or "universe_domain" not in data
            or data["type"] != "service_account"
            or data["universe_domain"] != "googleapis.com"
        ):
            raise ValueError(INVALID_AUTHENTICATION_KEY)

        if (
            not data.get("client_email")
            or not data["client_email"].startswith(settings.FABUBLOX_SA_EMAIL_PREFIX)
            or "@" not in data["client_email"]
        ):
            raise ValueError(INVALID_AUTHENTICATION_KEY)

        return {
            "client_email": data["client_email"],
            "key_instance_id": data["client_email"].split("@")[0].replace(settings.FABUBLOX_SA_EMAIL_PREFIX, ""),
        }

    def set_configuration_instance_id(self):
        configuration = FabubloxIntegrationConfiguration.get_instance()
        if not configuration.instance_id and self.key_instance_id:
            # set the instance_id if not already set
            configuration.instance_id = self.key_instance_id
            configuration.save()


class SingletonModel(models.Model):
    """
    Abstract base class for singleton models.
    """

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.pk = 1
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        pass

    @classmethod
    def get_instance(cls):
        instance, _ = cls.objects.get_or_create(pk=1)
        return instance


class ToolField(models.TextChoices):
    NAME = "name", "Name"
    DESCRIPTION = "description", "Description"
    CATEGORY = "category", "Category"
    LOCATION = "location", "Location"
    PRIMARY_OWNER = "primary_owner", "Primary Owner"
    PRIMARY_OWNER_FIRST_NAME = "primary_owner_first_name", "Primary Owner First Name"
    PRIMARY_OWNER_LAST_NAME = "primary_owner_last_name", "Primary Owner Last Name"
    PRIMARY_OWNER_EMAIL = "primary_owner_email", "Primary Owner Email"
    BACKUP_OWNERS = "backup_owners", "Backup Owners"
    BACKUP_OWNERS_FIRST_NAME = "backup_owners_first_name", "Backup Owners First Name"
    BACKUP_OWNERS_LAST_NAME = "backup_owners_last_name", "Backup Owners Last Name"
    BACKUP_OWNERS_EMAIL = "backup_owners_email", "Backup Owners Email"

    @classmethod
    def get_all_fields(cls):
        return list(cls.values)


class FabubloxIntegrationConfiguration(SingletonModel):
    instance_id = models.CharField(
        max_length=255,
        unique=True,
        blank=True,
        null=True,
        help_text="Unique identifier for this NEMO instance in Fabublox (Do not change this unless instructed by Fabublox support).",
    )
    data_sync_bucket_name = models.CharField(
        max_length=255,
        unique=True,
        blank=True,
        null=True,
        help_text="Name of the bucket used for data synchronization with Fabublox (Do not change this unless instructed by Fabublox support).",
    )
    admin_permission = MultiRoleGroupPermissionChoiceField(
        null=False,
        blank=False,
        groups=True,
        default="is_superuser",
        help_text="The roles/groups required for users to access the Fabublox integration features.",
    )
    shared_tool_fields = models.JSONField(
        default=ToolField.get_all_fields,
        blank=True,
        null=True,
        help_text="List of fields to be shared with Fabublox for tool synchronization.",
    )

    def __str__(self):
        return f"Fabublox Integration Configuration"

    @classmethod
    def get_fabublox_authentication_key(cls):
        key_location = getattr(settings, "FABUBLOX_AUTHENTICATION_KEY_LOCATION", None)
        if key_location:
            return FabubloxAuthenticationKey(key_location)
        return None

    def key_health_check(self):
        try:
            authentication_key = self.get_fabublox_authentication_key()
            if not authentication_key:
                return False, NOTFOUND_AUTHENTICATION_KEY
            return True, "Authentication key is valid."
        except ValueError as e:
            return False, e

    def check_readiness(self):
        key_health_pass, message = self.key_health_check()
        if not key_health_pass:
            raise ValueError(message)

        if not self.instance_id or not self.instance_id.strip():
            raise ValueError("Instance ID is not set or is invalid.")

        if not self.data_sync_bucket_name or not self.data_sync_bucket_name.strip():
            raise ValueError("Data synchronization bucket name is not set or is invalid.")

    @classmethod
    def get_admin_permission_field(cls) -> MultiRoleGroupPermissionChoiceField:
        return cls._meta.get_field("admin_permission")

    def has_access(self, user):
        return self.get_admin_permission_field().has_user_roles(self.admin_permission, user)

    def clean(self):
        errors = {}

        if self.shared_tool_fields:
            if not isinstance(self.shared_tool_fields, list):
                errors["share_tool_fields"] = "Shared tool fields must be a list."
            else:
                fields = self.shared_tool_fields or []
                for field in fields:
                    if field not in ToolField.values:
                        errors["share_tool_fields"] = f"Invalid shared tool field: {field}"

                primary_owner_fields = {
                    ToolField.PRIMARY_OWNER_FIRST_NAME,
                    ToolField.PRIMARY_OWNER_LAST_NAME,
                    ToolField.PRIMARY_OWNER_EMAIL,
                }
                backup_owners_fields = {
                    ToolField.BACKUP_OWNERS_FIRST_NAME,
                    ToolField.BACKUP_OWNERS_LAST_NAME,
                    ToolField.BACKUP_OWNERS_EMAIL,
                }
                intersect_primary_owner = primary_owner_fields & set(fields)
                if intersect_primary_owner and ToolField.PRIMARY_OWNER not in fields:
                    fields.append(ToolField.PRIMARY_OWNER)
                elif not intersect_primary_owner and ToolField.PRIMARY_OWNER in fields:
                    fields.remove(ToolField.PRIMARY_OWNER)
                intersect_backup_owners = backup_owners_fields & set(fields)
                if intersect_backup_owners and ToolField.BACKUP_OWNERS not in fields:
                    fields.append(ToolField.BACKUP_OWNERS)
                elif not intersect_backup_owners and ToolField.BACKUP_OWNERS in fields:
                    fields.remove(ToolField.BACKUP_OWNERS)

                self.shared_tool_fields = fields

        if len(errors) > 0:
            raise ValidationError(errors)


class FabubloxDataSyncStatus(models.Model):
    class Meta:
        unique_together = ("content_type", "object_id")
        indexes = [
            models.Index(fields=["content_type", "object_id", "-last_synced"]),
            models.Index(fields=["-last_synced"]),
        ]

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, help_text="Content type of the model being synced"
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    last_synced = models.DateTimeField(auto_now=True, help_text="Timestamp of the last sync attempt with FabuBlox")
    last_synced_by = ForeignKey(
        User,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="fabublox_sync_status",
        help_text="User who last synced this entity",
    )


class JobStatus(object):
    PENDING = 0
    IN_PROGRESS = 1
    COMPLETED = 2
    PARTIALLY_COMPLETED = 3
    FAILED = 4
    Choices = (
        (PENDING, "Pending"),
        (IN_PROGRESS, "In Progress"),
        (COMPLETED, "Completed"),
        (PARTIALLY_COMPLETED, "Partially Completed"),
        (FAILED, "Failed"),
    )


class FabubloxSyncFailure(models.Model):
    job = models.ForeignKey(
        "FabubloxSynchronizationJob", on_delete=models.CASCADE, related_name="failures", help_text="Synchronization job"
    )
    item_id = models.PositiveIntegerField(help_text="ID of the item that failed to sync")
    item_name = models.CharField(max_length=255, help_text="Name of the item that failed")
    error_message = models.TextField(help_text="Error message from the failed sync attempt")
    failed_at = models.DateTimeField(auto_now_add=True, help_text="Timestamp when the sync failed")

    class Meta:
        indexes = [
            models.Index(fields=["job", "-failed_at"]),
        ]


class FabubloxSynchronizationJob(models.Model):
    class Meta:
        indexes = [
            models.Index(fields=["status", "-created_at"]),
            models.Index(fields=["-created_at"]),
        ]

    status = models.IntegerField(
        choices=JobStatus.Choices, default=JobStatus.PENDING, help_text="Current status of the synchronization job"
    )
    total_items = models.PositiveIntegerField(default=0, help_text="Total number of items to be synchronized")
    succeeded_items = models.PositiveIntegerField(default=0, help_text="Number of items successfully synchronized")
    failed_items = models.PositiveIntegerField(default=0, help_text="Number of items that failed to synchronize")
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, help_text="Content type of the model being synced"
    )
    started_by = ForeignKey(
        User,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="fabublox_sync_jobs",
        help_text="User who started this synchronization job",
    )
    item_ids = models.JSONField(help_text="List of item IDs being synchronized")
    created_at = models.DateTimeField(auto_now_add=True, help_text="Timestamp when the job was created")
    updated_at = models.DateTimeField(auto_now=True, help_text="Timestamp when the job was last updated")

    @property
    def processed_items(self):
        return self.succeeded_items + self.failed_items

    @property
    def progress_percent(self):
        if self.total_items == 0:
            return 0
        return round((self.processed_items / self.total_items) * 100)

    @property
    def failed_items_details(self):
        return [{"item": failure.item_name, "error": failure.error_message} for failure in self.failures.all()]

    @property
    def is_stale(self):
        """Check if job appears to be abandoned (thread died/server restarted)."""
        from datetime import timedelta

        if self.status not in [JobStatus.IN_PROGRESS, JobStatus.PENDING]:
            return False
        timeout_minutes = getattr(settings, "FABUBLOX_JOB_TIMEOUT_MINUTES", 5)
        return self.updated_at < now() - timedelta(minutes=timeout_minutes)

    def handle_interrupted(self):
        """Update job status if it was likely interrupted (thread died, server restart, etc)."""
        with transaction.atomic():
            job = type(self).objects.select_for_update().get(pk=self.pk)

            if job.succeeded_items == job.total_items:
                job.status = JobStatus.COMPLETED
            elif job.failed_items == job.total_items:
                job.status = JobStatus.FAILED
            elif job.succeeded_items + job.failed_items == job.total_items:
                job.status = JobStatus.PARTIALLY_COMPLETED
            else:
                # likely interrupted, mark as failed
                job.status = JobStatus.FAILED

            job.save(update_fields=["status"])

    @classmethod
    def create_job(cls, content_type, item_ids, started_by):
        return cls.objects.create(
            content_type=content_type,
            started_by=started_by,
            item_ids=list(item_ids),
            total_items=len(item_ids),
            status=JobStatus.PENDING,
        )

    @classmethod
    def get_in_progress_jobs(cls):
        return cls.objects.filter(status__in=[JobStatus.IN_PROGRESS, JobStatus.PENDING]).order_by("-created_at")

    @classmethod
    def get_history(cls):
        return cls.objects.filter(
            status__in=[JobStatus.COMPLETED, JobStatus.PARTIALLY_COMPLETED, JobStatus.FAILED]
        ).order_by("-created_at")

    @classmethod
    def cleanup_stale_jobs(cls):
        """
        Find and mark stale jobs as failed.

        This method is throttled to run at most once per configured interval
        to avoid overwhelming the database during frequent polling.
        """
        global _last_cleanup_time

        cleanup_interval_minutes = getattr(settings, "FABUBLOX_JOB_CLEANUP_INTERVAL_MINUTES", 2)
        cleanup_interval = timedelta(minutes=cleanup_interval_minutes)

        current_time = now()

        if _last_cleanup_time is None or (current_time - _last_cleanup_time) >= cleanup_interval:
            _last_cleanup_time = current_time

            in_progress_jobs = cls.get_in_progress_jobs()
            cleaned = 0
            for job in in_progress_jobs:
                if job.is_stale:
                    job.handle_interrupted()
                    cleaned += 1
            return cleaned

        return 0



from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.contenttypes.models import ContentType
from django.db.models import OuterRef, Subquery
from django.http import HttpResponseNotFound
from django.shortcuts import render, redirect
from django.views.decorators.http import require_http_methods, require_GET, require_POST
from NEMO.models import Tool
from NEMO.views.pagination import SortedPaginator

from NEMO_fabublox_integration.forms import FabubloxIntegrationSharedToolFieldsForm
from NEMO_fabublox_integration.models import (
    FabubloxIntegrationConfiguration,
    FabubloxSynchronizationJob,
    FabubloxDataSyncStatus,
)
from NEMO_fabublox_integration.serializers import get_tool_serializer
from NEMO_fabublox_integration.synchronize import synchronize_job


def can_access_fabublox_integration(user):
    return FabubloxIntegrationConfiguration.get_instance().has_access(user)


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_GET
def configuration(request):
    configuration_obj = FabubloxIntegrationConfiguration.get_instance()
    key_health_pass, key_health_message = configuration_obj.key_health_check()

    return render(
        request,
        "NEMO_fabublox_integration/configuration.html",
        {
            "key_health_message": key_health_message,
            "key_health_pass": key_health_pass,
            "shared_tool_fields": configuration_obj.shared_tool_fields,
        },
    )


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_POST
def update_shared_tool_fields(request):
    configuration_obj = FabubloxIntegrationConfiguration.get_instance()
    fields = request.POST.getlist("shared_tool_fields", [])
    shared_tool_fields_form = FabubloxIntegrationSharedToolFieldsForm(
        {"shared_tool_fields": fields}, instance=configuration_obj
    )
    if shared_tool_fields_form.is_valid():
        shared_tool_fields_form.save()
        messages.success(request, "Shared tool fields updated successfully.")
        return redirect("fabublox_configuration")

    configuration_obj.refresh_from_db()
    key_health_pass, key_health_message = configuration_obj.key_health_check()

    return render(
        request,
        "NEMO_fabublox_integration/configuration.html",
        {
            "key_health_message": key_health_message,
            "key_health_pass": key_health_pass,
            "shared_tool_fields": configuration_obj.shared_tool_fields,
        },
    )


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_GET
def tools(request):
    configuration_obj = FabubloxIntegrationConfiguration.get_instance()
    try:
        configuration_obj.check_readiness()
        search = request.GET.get("search", None)
        if search:
            tools_queryset = Tool.objects.filter(name__icontains=search)
        else:
            tools_queryset = Tool.objects.all()
        number_of_tools = Tool.objects.count()
        tool_content_type = ContentType.objects.get_for_model(Tool)
        sync_status_qs = FabubloxDataSyncStatus.objects.filter(
            content_type=tool_content_type, object_id=OuterRef("pk")
        ).order_by("-last_synced")
        tools_queryset = tools_queryset.annotate(
            last_synced=Subquery(sync_status_qs.values("last_synced")[:1]),
            last_synced_by=Subquery(sync_status_qs.values("last_synced_by__username")[:1]),
        )
        page = SortedPaginator(tools_queryset, request, order_by="name").get_current_page()
        return render(
            request,
            "NEMO_fabublox_integration/tools.html",
            {"page": page, "search_term": search, "number_of_tools": number_of_tools, "ready": True},
        )
    except ValueError as e:
        return render(request, "NEMO_fabublox_integration/tools.html", {"ready": False, "configuration_readiness": e})


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_GET
def synchronizations(request):
    # Clean up any stale jobs (from server restarts, thread crashes, etc)
    FabubloxSynchronizationJob.cleanup_stale_jobs()

    in_progress = FabubloxSynchronizationJob.get_in_progress_jobs()
    others = FabubloxSynchronizationJob.get_history()
    page = SortedPaginator(others, request, order_by="-created_at").get_current_page()
    return render(
        request,
        "NEMO_fabublox_integration/synchronization.html",
        {"page": page, "in_progress": in_progress, "has_in_progress": in_progress.exists()},
    )


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_http_methods(["GET"])
def in_progress_synchronizations(request):
    # Clean up any stale jobs (from server restarts, thread crashes, etc)
    FabubloxSynchronizationJob.cleanup_stale_jobs()

    in_progress = FabubloxSynchronizationJob.get_in_progress_jobs()
    if not in_progress.exists():
        return HttpResponseNotFound("No synchronization in progress.")

    return render(
        request,
        "NEMO_fabublox_integration/in_progress_synchronization.html",
        {"in_progress": in_progress},
    )


@login_required
@user_passes_test(can_access_fabublox_integration)
@require_POST
def synchronize_tools(request):
    configuration_obj = FabubloxIntegrationConfiguration.get_instance()
    user = request.user

    try:
        configuration_obj.check_readiness()
    except Exception as e:
        messages.error(
            request,
            f"Your FabuBlox integration configuration is incomplete or invalid: {e}",
            "data-speed=40000 data-trigger=manual",
        )
        return redirect("fabublox_tools")

    # Clean up any stale jobs before checking for in-progress jobs
    FabubloxSynchronizationJob.cleanup_stale_jobs()
    tool_content_type = ContentType.objects.get_for_model(Tool)

    in_progress = FabubloxSynchronizationJob.get_in_progress_jobs().filter(content_type=tool_content_type)

    # Only allow one synchronization job at a time
    if in_progress.exists():
        messages.error(
            request,
            "A synchronization job is already in progress. Please wait for it to complete.",
            "data-speed=40000 data-trigger=manual",
        )
        return redirect("fabublox_tools")

    sync_all = request.POST.get("sync_all", "false").lower() == "true"

    try:
        if sync_all:
            tool_ids = list(Tool.objects.values_list("id", flat=True))
        else:
            tool_id_strings = request.POST.getlist("tool_id")
            tool_ids = []
            for tool_id_str in tool_id_strings:
                try:
                    tool_ids.append(int(tool_id_str))
                except (ValueError, TypeError):
                    messages.error(request, f"Invalid tool ID: {tool_id_str}", "data-speed=40000 data-trigger=manual")
                    return redirect("fabublox_tools")

        if not tool_ids:
            messages.error(request, "No tools selected for synchronization.", "data-speed=40000 data-trigger=manual")
            return redirect("fabublox_tools")

        synchronize_job(get_tool_serializer(), tool_content_type, tool_ids, user)
        messages.success(request, f"Synchronization job started successfully for {len(tool_ids)} tool(s).")
        return redirect("fabublox_synchronization")

    except ValueError as e:
        messages.error(request, f"Failed to start synchronization job: {e}", "data-speed=40000 data-trigger=manual")
        return redirect("fabublox_tools")
    except Exception as e:
        messages.error(
            request, f"Unexpected error starting synchronization job: {e}", "data-speed=40000 data-trigger=manual"
        )
        return redirect("fabublox_tools")
